"""
Utility functions for testing and normalising scalar-ish data types

@author: David Megginson
@organization: UNOCHA
@license: Public Domain
@date: Started April 2018
@see: U{http://hxlstandard.org}
"""

import collections, datetime, dateutil.parser, re, six, unidecode


TOKEN_PATTERN = r'[A-Za-z][_0-9A-Za-z]*'
"""Regular-expression pattern for a single token."""

WHITESPACE_PATTERN = re.compile('\s+', re.MULTILINE)
"""Regular-expression pattern for multi-line whitespace."""

ISO_DATE_PATTERN = re.compile('^(?P<year>[12]\d\d\d)(?:Q(?P<quarter>[1-4])|W(?P<week>\d\d?)|-(?P<month>\d\d?)(?:-(?P<day>\d\d?))?)?$', re.IGNORECASE)
"""Regular expression for basic ISO 8601 dates, plus extension to recognise quarters."""

DEFAULT_DATE_1 = datetime.datetime(2015, 1, 1)

DEFAULT_DATE_2 = datetime.datetime(2016, 3, 3)

def normalise(s, col=None):
    """Intelligently normalise a value, optionally using the HXL hashtag for hints"""
    # TODO add lat/lon
    if col and col.tag == '#date' and is_date(s):
        return normalise_date(s)
    elif is_number(s):
        return normalise_number(s)
    else:
        return normalise_string(s)

def typeof(s, col=None):
    """Determine the type of a column value"""
    if col and col.tag == '#date' and is_date(s):
        return 'date'
    elif is_number(s):
        return 'number'
    elif is_empty(s):
        return 'empty'
    else:
        return 'string'
    
def flatten(value, is_subitem=False):
    """Flatten potential lists and dictionaries"""

    # keep it simple for now
    if value is None:
        return ''
    else:
        return str(value)

    
def is_truthy(s):
    """Check for a boolean-type true value
    @param s: the value to test
    @returns: True if the value is truthy
    """
    return normalise_string(s) in ['y', 'yes', 't', 'true', '1']


def is_empty(s):
    """Is this an empty value?
    None or whitespace only counts as empty; anything else doesn't.
    @param s: value to test
    @return: True if the value is empty
    """
    return (s is None or s == '' or str(s).isspace())


def is_string(v):
    """Test if a value is currently a string
    @param v: the value to test
    @returns: True if the value is a string
    """
    return isinstance(v, six.string_types)
    
def normalise_space(s):
    """Normalise whitespace only
    @param v: value to normalise
    @returns: string value with whitespace normalised
    """
    if is_empty(s):
        return ''
    else:
        s = str(s).strip().replace("\n", " ")
        return re.sub(
            WHITESPACE_PATTERN,
            ' ',
            s
        )

def normalise_string(s):
    """Normalise a string.
    Remove all leading and trailing whitespace. Convert to lower
    case. Replace all internal whitespace (including lineends) with a single space. Replace None with ''.
    @param s: the string to normalise
    @returns: the normalised string
    """
    if s is None:
        s = ''
    else:
        s = str(s)
    return normalise_space(unidecode.unidecode(s)).lower()


def is_number(v):
    """Test if a value contains something recognisable as a number.
    @param v: the value (string, int, float, etc) to test
    @returns: True if usable as a number
    @see: L{normalise_number}
    """
    try:
        float(v)
        return True
    except:
        return False

def normalise_number(v):
    """Attempt to convert a value to a number.
    Will convert to int type if it has no decimal places.
    @param v: the value (string, int, float, etc) to convert.
    @returns: an int or float value
    @exception ValueError: if the value cannot be converted
    @see: L{is_number}
    """
    n = float(v)
    if n == int(n):
        return int(n)
    else:
        return n

def is_date(v):
    """Test if a value contains something recognisable as a date.
    @param v: the value (string, etc) to test
    @returns: True if usable as a date
    @see: L{normalise_date}
    """
    v = normalise_space(v)
    result = ISO_DATE_PATTERN.match(v)
    if result:
        # lots of ugly ISO date hackery
        # TODO not distinguishing non-leap-years yet

        def in_range(n, min, max):
            if n is None:
                return True
            else:
                n = int(n)
                return (n >= min and n <= max)

        if not in_range(result.group('year'), 1900, 2100):
            return False
        if not in_range(result.group('month'), 1, 12):
            return False
        if result.group('day') is not None:
            month = int(result.group('month'))
            if month == 2 and not in_range(result.group('day'), 1, 29):
                return False
            elif month in [4, 6, 9, 11] and not in_range(result.group('day'), 1, 30):
                return False
            elif not in_range(result.group('day'), 1, 31):
                return False
        if not in_range(result.group('quarter'), 1, 4):
            return False
        if not in_range(result.group('week'), 0, 53):
            return False

        return True
    try:
        result = dateutil.parser.parse(v)
        return True
    except ValueError as e:
        return False

def normalise_date(v):
    """Normalise a string as a date.
    @param s: the string to normalise as a date
    @returns: the date in ISO 8601 format or quarters (extension)
    @exception ValueError: if value cannot be parsed as a date
    @see: L{is_date}
    """

    def make_date(year, quarter=None, month=None, week=None, day=None):
        if quarter:
            # *not* real ISO 8601
            return '{:04d}Q{:01d}'.format(int(year), int(quarter))
        elif week:
            return '{:04d}W{:02d}'.format(int(year), int(week))
        elif month:
            if day:
                return '{:04d}-{:02d}-{:02d}'.format(int(year), int(month), int(day))
            else:
                return '{:04d}-{:02d}'.format(int(year), int(month))
        else:
            return '{:04d}'.format(int(year))
        

    # First, try our quick ISO date pattern, extended to support quarter notation
    v = normalise_space(v)
    result = ISO_DATE_PATTERN.match(v)
    if result:
        return make_date(
            result.group('year'),
            quarter=result.group('quarter'),
            month=result.group('month'),
            week=result.group('week'),
            day=result.group('day')
        )

    # revert to full date parsing
    # we parse the date twice, to detect any default values Python might have filled in
    date1 = dateutil.parser.parse(v, default=DEFAULT_DATE_1)
    date2 = dateutil.parser.parse(v, default=DEFAULT_DATE_2)
    return make_date(
        date1.year,
        month=(date1.month if date1.month==date2.month else None),
        day=(date1.day if date1.day==date2.day else None)
    )
    

def is_list(e):
    """Test if a value is a Python sequence (other than a string)
    @param e: the value to test
    @return: True if the value is a sequence; False otherwise"""
    if not isinstance(e, collections.Sequence) or isinstance(e, six.string_types):
        return False
    else:
        return True




import hxl, io, xml.sax

class SAXHandler(xml.sax.handler.ContentHandler):
    """SAX event handler to convert an IATI activity report to a series of CSV-style rows"""

    OUTPUT_SPEC = [
        ['IATI id', '#activity+id+v_iati_activities', 1],
        ['Last updated', '#date+updated', 1],
        ['Activity status', '#status', 1],
        ['Funding org', '#org+name+funder', 1],
        ['Programming org', '#org+name+prog', 1],
        ['Implementing partner', '#org+name+impl', 1],
        ['Activity', '#activity+name', 1],
        ['DAC sector', '#sector+name', 3],
        ['Country code', '#country+code+recipient', 3],
        ['Planned start date', '#date+planned+start', 1],
        ['Actual start date', '#date+actual+start', 1],
        ['Planned end date', '#date+planned+end', 1],
        ['Actual end date', '#date+actual+end', 1],
    ]
    """Specification for what data should appear in the CSV"""

    STATUS_CODES = {
        '1': 'Pipeline/Identification',
        '2': 'Implementation',
        '3': 'Completion',
        '4': 'Post-completion',
        '5': 'Cancelled',
        '6': 'Suspended'
    }
    """Lookup table for IATI status codes"""

    DAC_SECTOR_CODES = {
        "110": "Education",
        "111": "Education, Level Unspecified",
        "112": "Basic Education",
        "113": "Secondary Education",
        "114": "Post-Secondary Education",
        "120": "Health",
        "121": "Health, General",
        "122": "Basic Health",
        "130": "Population Policies/Programmes & Reproductive Health",
        "140": "Water Supply & Sanitation",
        "150": "Government & Civil Society",
        "151": "Government & Civil Society-general",
        "152": "Conflict, Peace & Security",
        "160": "Other Social Infrastructure & Services",
        "210": "Transport & Storage",
        "220": "Communications",
        "230": "Energy",
        "231": "Energy Policy",
        "232": "Energy generation, renewable sources",
        "233": "Energy generation, non-renewable sources",
        "234": "Hybrid energy plants",
        "235": "Nuclear energy plants",
        "236": "Energy distribution",
        "240": "Banking & Financial Services",
        "250": "Business & Other Services",
        "310": "Agriculture, Forestry, Fishing",
        "311": "Agriculture",
        "312": "Forestry",
        "313": "Fishing",
        "320": "Industry, Mining, Construction",
        "321": "Industry",
        "322": "Mineral Resources & Mining",
        "323": "Construction",
        "331": "Trade Policies & Regulations",
        "332": "Tourism",
        "410": "General Environment Protection",
        "430": "Other Multisector",
        "510": "General Budget Support",
        "520": "Developmental Food Aid/Food Security Assistance",
        "530": "Other Commodity Assistance",
        "600": "Action Relating to Debt",
        "720": "Emergency Response",
        "730": "Reconstruction Relief & Rehabilitation",
        "740": "Disaster Prevention & Preparedness",
        "910": "Administrative Costs of Donors",
        "930": "Refugees in Donor Countries",
        "998": "Unallocated / Unspecified"
    }
    """OECD DAC high-level sector codes"""

    def __init__ (self):
        """Create a new SAX event handler for IATI->CSV conversion.
        """
        super().__init__()

        self.element_stack = list()
        self.activity = dict()
        self.content = ''
        self.rows = list()

    def startElement(self, name, atts):
        """Beginning of a nested XML element"""
        super().startElement(name, atts)

        # Trigger actions based on the element name
        if name == 'iati-activities':
            self.write_headers()
            self.write_hashtags()
        elif name == 'iati-activity':
            self.activity = {}
            self.add_prop('#date+updated', atts.get('last-updated-datetime')[:10])
        elif name == 'reporting-org':
            self.add_prop('#org+code+reporting', atts.get('ref'))
        elif name == 'participating-org':
            role = atts.get('role')
            if role == '1':
                self.add_prop('#org+code+funder', atts.get('ref'))
            elif role in ('2', '3'):
                self.add_prop('#org+code+prog', atts.get('ref'))
            else:
                self.add_prop('#org+code+impl', atts.get('ref'))
        elif name == 'recipient-country':
            self.add_prop('#country+code+recipient', atts.get('code'))
        elif name == 'sector':
            code = atts.get('code')
            if code:
                sector = self.DAC_SECTOR_CODES.get(code[:3])
                self.add_prop('#sector+name', sector)
                self.add_prop('#sector+code+v_iati_sectors', code)
        elif name == 'activity-date':
            date = atts.get('iso-date')
            type = atts.get('type')
            if type == '1':
                self.add_prop('#date+planned+start', date)
            elif type == '2':
                self.add_prop('#date+actual+start', date)
            elif type == '3':
                self.add_prop('#date+planned+end', date)
            elif type == '4':
                self.add_prop('#date+actual+end', date)
        elif name == 'activity-status':
            if self.STATUS_CODES.get(atts.get('code')):
                self.add_prop('#status', self.STATUS_CODES[atts['code']])
            
        self.element_stack.append((name, atts,))
        self.content = ''

    def endElement(self, name):
        """End of a nested XML element"""
        super().endElement(name)
        
        self.element_stack.pop()

        # Trigger actions based on the element name
        if name == 'iati-activity':
            self.write_data()
        elif name == 'iati-identifier':
            self.add_prop('#activity+id+v_iati_activities', self.content)
        elif name == 'narrative':
            if self.has_parent('title'):
                self.add_prop('#activity+name', self.content)
            elif self.has_parent('reporting-org'):
                self.add_prop('#org+name+reporting', self.content)
            elif self.has_parent('participating-org'):
                role = self.get_parent_att('role')
                if role == '1':
                    self.add_prop('#org+name+funder', self.content)
                elif role == '2' or role == '3':
                    self.add_prop('#org+name+prog', self.content)
                elif role == '4':
                    self.add_prop('#org+name+impl', self.content)

    def characters(self, content):
        """Any chunk of character data (may not be complete)"""
        super().characters(content)

        # Add to the current text content (will be used by endElement)
        self.content += content

    def add_prop(self, hashtag, value):
        """Add a property to the current row (supports multiple values)"""
        if not self.activity.get(hashtag):
            self.activity[hashtag] = []
        self.activity[hashtag].append(value)

    def has_parent(self, name):
        """Check for a specific parent element"""
        return (self.element_stack[-1][0] == name)

    def get_parent_att(self, att_name):
        return self.element_stack[-1][1].get(att_name)

    def write_headers(self):
        """Write the CSV header row"""
        row = []
        for spec in self.OUTPUT_SPEC:
            for i in range(0, spec[2]):
                row.append(spec[0])
        self.rows.append(row)

    def write_hashtags(self):
        """Write the CSV hashtag row"""
        row = []
        for spec in self.OUTPUT_SPEC:
            for i in range(0, spec[2]):
                row.append(spec[1].replace('+', ' +'))
        self.rows.append(row)

    def write_data(self):
        """Write the current activity as a CSV data row"""
        row = []
        for spec in self.OUTPUT_SPEC:
            values = self.activity.get(spec[1])
            # special kludge
            if values and (spec[1] == '#sector+name' or spec[1] == '#org+name+participating'):
                values = list(dict.fromkeys(values))
            for i in range(0, spec[2]):
                if values and len(values) > i and values[i]:
                    row.append(values[i])
                else:
                    row.append('')
        self.rows.append(row)

